import versioneer
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

requires = [
    "anndata>=0.7.1",
    "Cython",
    "docopt",
    "fisher>=0.1.9",
    "forceatlas2-python",
    "harmony-pytorch",
    "hnswlib",
    "joblib",
    "lightgbm==2.2.1",
    "loompy",
    "louvain-github",
    "matplotlib>=2.0.0",
    "MulticoreTSNE-modified",
    "natsort",
    "numba",
    "numpy",
    "pandas>=0.24",
    "plotly",
    "pyarrow",
    "pybind11",
    "python-igraph",
    "scikit-learn>=0.21.3",
    "scikit-misc",
    "scipy",
    "scplot",
    "seaborn",
    "setuptools",
    "statsmodels",
    "tables",
    "torch",
    "umap-learn>=0.3.9",
    "xlsxwriter"
]

setup(
    name="pegasuspy",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Pegasus is a Python package for analyzing sc/snRNA-seq data of millions of cells",
    long_description=long_description,
    url="https://github.com/klarman-cell-observatory/pegasus",
    author="Yiming Yang, Joshua Gould and Bo Li",
    author_email="cumulus-support@googlegroups.com",
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Framework :: Jupyter",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="single cell/nucleus genomics analysis",
    packages=find_packages(),
    install_requires=requires,
    extras_require=dict(
        fitsne=["fitsne"],
        leiden=['leidenalg'],
        mkl=["mkl"]
    ),
    python_requires="~=3.5",
    package_data={
        "pegasus.annotate_cluster": [
            "human_immune_cell_markers.json",
            "mouse_immune_cell_markers.json",
            "mouse_brain_cell_markers.json",
            "human_brain_cell_markers.json",
        ],
        "pegasus.check_sample_indexes": ["chromium-dna-sample-indexes-plate.json"],
    },
    entry_points={"console_scripts": ["pegasus=pegasus.__main__:main"]},
)



import time

import numpy as np
import pandas as pd
from pegasus.io import read_input
from scipy.sparse import issparse


def scp_write_coords(data, output_name):
    cluster_labels = []
    for col_name in data.obs.columns:
        if col_name.find("labels") >= 0:
            cluster_labels.append(col_name)
    df_labels = data.obs[cluster_labels]
    clu_str = group_str = ""
    if len(cluster_labels) > 0:
        clu_str = "".join(["\t" + x for x in cluster_labels])
        group_str = "".join(["\tgroup"] * len(cluster_labels))

    for basis in data.obsm_keys():
        basis_X = data.obsm[basis]
        coords = ["X", "Y"] if basis_X.shape[1] == 2 else ["X", "Y", "Z"]
        coo_str = "\t".join(coords)
        num_str = "\t".join(["numeric"] * len(coords))
        coord_file = "{}.scp.{}.coords.txt".format(output_name, basis)
        with open(coord_file, "w") as fout:
            fout.write("NAME\t{coo}{clu}\n".format(coo=coo_str, clu=clu_str))
            fout.write("TYPE\t{coo}{clu}\n".format(coo=num_str, clu=group_str))
        df_out = pd.DataFrame(
            basis_X[:, 0: len(coords)],
            columns=coords,
            index=data.obs_names,
        )
        df_out = pd.concat([df_out, df_labels], axis=1)
        df_out.to_csv(coord_file, sep="\t", header=False, mode="a")
        print("Coordinate file {} is written.".format(coord_file))


def scp_write_metadata(data, output_name):
    ban = ["n_genes", "n_counts", "percent_mito", "pseudotime"]
    meta = []
    for col_name in data.obs.columns:
        if (col_name not in ban) and (col_name.find("labels") < 0):
            meta.append(col_name)
    meta_str = "".join(["\t" + x for x in meta])
    group_str = "".join(["\tgroup"] * len(meta))
    metadata_file = "{}.scp.metadata.txt".format(output_name)
    with open(metadata_file, "w") as fout:
        fout.write("NAME{meta}\n".format(meta=meta_str))
        fout.write("TYPE{meta}\n".format(meta=group_str))
    data.obs[meta].to_csv(metadata_file, sep="\t", header=False, mode="a")
    print("Metadata file {} is written.".format(metadata_file))


def write_market_matrix(mtx_file, X, round_to):
    with open(mtx_file, "w") as fout:
        fmt_str = "{{}} {{}} {{:.{}f}}\n".format(round_to)
        fout.write("%%MatrixMarket matrix coordinate real general\n%\n")
        if issparse(X):
            X = X.tocoo()
            fout.write("{} {} {}\n".format(X.shape[0], X.shape[1], (X.data != 0).sum()))
            for x, y, value in zip(X.row, X.col, X.data):
                if value != 0:
                    fout.write(fmt_str.format(x + 1, y + 1, value))
        else:
            fout.write("{} {} {}\n".format(X.shape[0], X.shape[1], np.count_nonzero(X)))
            for x, y in zip(*X.nonzero()):
                fout.write(fmt_str.format(x + 1, y + 1, X[x, y]))


def write_dense_matrix(expr_file, data, round_to):
    fmt_str = "{{:.{}f}}".format(round_to)
    exprmat = data.X.transpose()
    if issparse(exprmat):
        exprmat = exprmat.toarray()
    with open(expr_file, "w") as fout:
        fout.write("GENE\t" + "\t".join(data.obs_names) + "\n")
        for i in range(exprmat.shape[0]):
            fout.write(
                data.var_names[i]
                + "\t"
                + "\t".join(
                    [fmt_str.format(x) if x > 0.0 else "0" for x in exprmat[i, :]]
                )
                + "\n"
            )
    print("Expression matrix {} is written.".format(expr_file))


def scp_write_expression(data, output_name, is_sparse=True, round_to=2):

    if is_sparse:
        barcode_file = "{}.scp.barcodes.tsv".format(output_name)
        with open(barcode_file, "w") as fout:
            fout.write("\n".join(data.obs_names) + "\n")
        print("Barcode file {} is written.".format(barcode_file))
        gene_file = "{}.scp.features.tsv".format(output_name)
        df = pd.DataFrame(
            {"gene_names": data.var_names, "gene_ids": data.var["gene_ids"]}
        )[["gene_ids", "gene_names"]]
        with open(gene_file, "w") as fout:
            df.to_csv(fout, sep=" ", header=False, index=False)
        print("Features file {} is written.".format(gene_file))
        mtx_file = "{}.scp.matrix.mtx".format(output_name)
        write_market_matrix(mtx_file, data.X.transpose().tocsr(), round_to)
        print("Matrix file {} is written.".format(mtx_file))
    else:
        expr_file = "{}.scp.expr.txt".format(output_name)
        write_dense_matrix(expr_file, data, round_to)


def run_scp_output(
    input_h5ad_file: str, output_name: str, is_sparse: bool = True, round_to: int = 2
):
    """Generate outputs for single cell portal.

    Parameters
    ----------
    input_h5ad_file: ``str``
        Input h5ad file name.

    output_name: ``str``
        Name prefix for output files.

    is_sparse: ``bool``, optional, default: ``True``
        If ``True``, enforce the count matrix to be sparse after written into files.

    round_to: ``int``, optional, default: ``2``
        Round numbers to ``round_to`` decimal places.

    Returns
    -------
    ``None``

    Generate several files:
        * ``output_name.scp.basis.coords.txt``, where ``basis`` is for each key in ``adata.obsm`` field.
        * ``output_name.scp.metadata.txt``.
        * Gene expression files:
            * If in sparse format:
                * ``output_name.scp.features.tsv``, information on genes;
                * ``output_name.scp.barcodes.tsv``, information on cell barcodes;
                * ``output_name.scp.matrix.mtx``, count matrix.
            * If not in sparse:
                * ``output_name.scp.expr.txt``.

    Examples
    --------
    >>> pg.run_scp_output("result.h5ad", output_name = "scp_result")
    """
    adata = read_input(input_h5ad_file, h5ad_mode="a")
    start = time.perf_counter()
    scp_write_coords(adata, output_name)
    scp_write_metadata(adata, output_name)
    scp_write_expression(adata, output_name, is_sparse, round_to)
    end = time.perf_counter()
    print("Time spent for generating SCP outputs is {:.2f}s.".format(end - start))



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from scipy.stats import zscore
from sklearn.metrics import adjusted_mutual_info_score
from natsort import natsorted
from matplotlib import rcParams
from pandas.api.types import is_numeric_dtype
from collections import namedtuple

from typing import List

from .plot_utils import get_palettes, transform_basis


Restriction = namedtuple("negation", "values")


class RestrictionParser:
    def __init__(self, restrictions: List[str]):
        self.restrs = {}
        for restr_str in restrictions:
            attr, value_str = restr_str.split(":")
            negation = False
            if value_str[0] == "~":
                negation = True
                value_str = value_str[1:]
            self.restr[attr] = Restriction(
                negation=negation, values=value_str.split(",")
            )

    def contains(self, attr: str) -> bool:
        return attr in self.restrs

    def get_attrs(self) -> List[str]:
        return self.restrs.keys()

    def get_satisfied(self, data: "AnnData") -> List[bool]:
        selected = np.ones(data.shape[0], dtype=bool)
        for attr, restr in self.restrs.items():
            labels = data.obs[attr].astype(str)
            if restr.negation:
                selected = selected & (~np.isin(labels, restr.values))
            else:
                selected = selected & np.isin(labels, restr.values)
        return selected

    def get_unsatisfied(self, data: "AnnData", apply_to_all: bool) -> List[bool]:
        unsel = np.zeros(data.shape[0], dtype=bool)
        if apply_to_all:
            for attr, restr in self.restrs.items():
                labels = data.obs[attr].astype(str)
                if restr.negation:
                    unsel = unsel | np.isin(labels, restr.values)
                else:
                    unsel = unsel | (~np.isin(labels, restr.values))
        return unsel

    def get_satisfied_per_attr(self, labels: List[str], attr: str) -> List[bool]:
        one_restr = self.restrs[attr]
        if one_restr.negation:
            return ~np.isin(labels, rest_vec)
        else:
            return np.isin(labels, rest_vec)

    def get_unsatisfied_per_attr(self, labels: List[str], attr: str) -> List[bool]:
        one_restr = self.restrs[attr]
        if one_restr.negation:
            return np.isin(labels, rest_vec)
        else:
            return ~np.isin(labels, rest_vec)


def get_nrows_and_ncols(num_figs, nrows, ncols):
    if nrows is None and ncols is None:
        nrows = int(np.sqrt(num_figs))
        ncols = (num_figs // nrows) + (num_figs % nrows > 0)
    elif nrows is None:
        nrows = (num_figs // ncols) + (num_figs % ncols > 0)
    elif ncols is None:
        ncols = (num_figs // nrows) + (num_figs % nrows > 0)

    return nrows, ncols


def set_up_kwargs(subplot_size, left, bottom, wspace, hspace):
    kwargs = {}
    if subplot_size is not None:
        kwargs["subplot_size"] = subplot_size
    if left is not None:
        kwargs["left"] = left
    if bottom is not None:
        kwargs["bottom"] = bottom
    if wspace is not None:
        kwargs["wspace"] = wspace
    if hspace is not None:
        kwargs["hspace"] = hspace

    return kwargs


def get_subplot_layouts(
    nrows=1,
    ncols=1,
    subplot_size=(4, 4),
    left=0.2,
    bottom=0.15,
    wspace=0.4,
    hspace=0.15,
    squeeze=True,
    sharex=True,
    sharey=True,
    frameon=False,
):
    left_margin = left * subplot_size[0]
    bottom_margin = bottom * subplot_size[1]
    right_space = wspace * subplot_size[0]
    top_space = hspace * subplot_size[1]

    figsize = (
        left_margin + subplot_size[0] * (1.0 + wspace) * ncols,
        bottom_margin + subplot_size[1] * (1.0 + hspace) * nrows,
    )
    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        figsize=figsize,
        squeeze=squeeze,
        sharex=sharex,
        sharey=sharey,
        frameon=frameon,
    )

    fig.subplots_adjust(
        left=left_margin / figsize[0],
        bottom=bottom_margin / figsize[1],
        right=1.0 - right_space / figsize[0],
        top=1.0 - top_space / figsize[1],
        wspace=wspace,
        hspace=hspace,
    )

    return fig, axes


def get_legend_ncol(label_size):
    return 1 if label_size <= 14 else (2 if label_size <= 30 else 3)


def as_category(labels):
    if isinstance(labels, pd.Series):
        labels = labels.values
    # if 'singlet' in np.unique(labels):
    #     return pd.Categorical(labels, categories = ['', 'singlet', 'doublet', 'unknown'])
    return pd.Categorical(labels, categories=natsorted(np.unique(labels)))


def get_marker_size(nsamples):
    return (240000.0 if nsamples > 300000 else 120000.0) / nsamples


def plot_composition(
    data,
    cluster,
    attr,
    style="frequency",
    stacked=True,
    logy=False,
    subplot_size=(6, 4),
    left=0.15,
    bottom=None,
    wspace=0.3,
    hspace=None,
    restrictions=[],
    **others,
):
    """Generate a composition plot, which shows the percentage of cells from each condition for every cluster.
    
    This function is used to generate composition plots, which are bar plots showing the cell compositions (from different conditions) for each cluster. This type of plots is useful to fast assess library quality and batch effects.

    Parameters
    ----------

    data : `anndata` object
        Single cell expression data as an anndata object.
    cluster : `str`
        A string represents cluster labels, e.g. louvain_labels.
    attr: `str`
        A sample attribute representing the condition, e.g. Donor.
    style: `str`, optional (default: `frequency`)
        Composition plot style. Can be either `frequency`, `count`, or 'normalized'. Within each cluster, the `frequency` style show the ratio of cells from each condition over all cells in the cluster, the `count` style just shows the number of cells from each condition, the `normalized` style shows the percentage of cells from the condition in this cluster over the total number of cells from the condition for each condition. 
    stacked: `bool`, optional (default: `True`)
        If stack the bars from each condition.
    logy: `bool`, optional (default: `False`)
        If show the y-axis in log10 scale
    subplot_size: `tuple`, optional (default: `(6, 4)`)
        The plot size (width, height) in inches.
    left: `float`, optional (default: `0.15`)
        This parameter sets the figure's left margin as a fraction of subplot's width (left * subplot_size[0]).
    bottom: `float`, optional (default: `0.15`)
        This parameter sets the figure's bottom margin as a fraction of subplot's height (bottom * subplot_size[1]),
    wspace: `float`, optional (default: `0.2`)
        This parameter sets the width between subplots and also the figure's right margin as a fraction of subplot's width (wspace * subplot_size[0]).
    hspace: `float`, optional (defualt: `0.15`)
        This parameter sets the height between subplots and also the figure's top margin as a fraction of subplot's height (hspace * subplot_size[1]).
    restrictions: `list[str]`, optional (default: `[]`)
        This parameter is used to select a subset of data to plot.
    
    Returns
    -------

    `Figure` object
        A `matplotlib.figure.Figure` object containing the composition plot.

    Examples
    --------
    >>> fig = plotting.plot_composition(data, 'louvain_labels', 'Donor', style = 'normalized', stacked = False)
    """

    kwargs = set_up_kwargs(subplot_size, left, bottom, wspace, hspace)
    fig, ax = get_subplot_layouts(**kwargs)

    restr_obj = RestrictionParser(restrictions)
    selected = restr_obj.get_satisfied(data)

    df = pd.crosstab(data.obs.loc[selected, cluster], data.obs.loc[selected, attr])
    df = df.reindex(
        index=natsorted(df.index.values), columns=natsorted(df.columns.values)
    )
    # df = df.reindex(index = natsorted(df.index.values), columns = ['singlet', 'doublet', 'unknown'])

    if style == "frequency":
        df = df.div(df.sum(axis=1), axis=0) * 100.0
    elif style == "normalized":
        df = df.div(df.sum(axis=0), axis=1) * 100.0

    palettes = get_palettes(df.shape[1])

    rot = None
    if len(max(df.index.astype(str), key=len)) < 5:
        rot = 0

    if logy and not stacked:
        df_sum = df.sum(axis=1)
        df_new = df.cumsum(axis=1)
        df_new = 10 ** df_new.div(df_sum, axis=0).mul(np.log10(df_sum), axis=0)
        df = df_new.diff(axis=1).fillna(value=df_new.iloc[:, 0:1], axis=1)
        df.plot(
            kind="bar",
            stacked=False,
            legend=False,
            logy=True,
            ylim=(1.01, df_sum.max() * 1.7),
            color=palettes,
            rot=rot,
            ax=ax,
        )
    else:
        df.plot(
            kind="bar",
            stacked=stacked,
            legend=False,
            logy=logy,
            color=palettes,
            rot=rot,
            ax=ax,
        )

    ax.grid(False)
    ax.set_xlabel("Cluster ID")
    ax.set_ylabel("Percentage" if style != "count" else "Count")
    ax.set_title(
        "AMI = {0:.4f}".format(
            adjusted_mutual_info_score(
                data.obs.loc[selected, cluster],
                data.obs.loc[selected, attr],
                average_method="arithmetic",
            )
        )
    )
    ax.legend(loc="center left", bbox_to_anchor=(1.05, 0.5))

    return fig


### Sample usage:
###    fig = plot_scatter(data, 'tsne', ['louvain_labels', 'hdbscan_labels_soft'], nrows = 1, ncols = 2, alpha = 0.5)
def plot_scatter(
    data,
    basis,
    attrs,
    restrictions=[],
    nrows=None,
    ncols=None,
    subplot_size=(4, 4),
    left=None,
    bottom=None,
    wspace=None,
    hspace=None,
    alpha=None,
    legend_fontsize=None,
    apply_to_all=True,
    show_background=False,
    vmin=0.0,
    vmax=1.0,
    **others,
):
    df = pd.DataFrame(
        data.obsm["X_" + basis][:, 0:2], columns=[basis + c for c in ["1", "2"]]
    )
    basis = transform_basis(basis)

    nattrs = len(attrs)
    nrows, ncols = get_nrows_and_ncols(nattrs, nrows, ncols)
    marker_size = get_marker_size(df.shape[0])

    kwargs = set_up_kwargs(subplot_size, left, bottom, wspace, hspace)
    fig, axes = get_subplot_layouts(nrows=nrows, ncols=ncols, squeeze=False, **kwargs)

    if legend_fontsize is None:
        legend_fontsize = rcParams["legend.fontsize"]

    restr_obj = RestrictionParser(restrictions)
    unsel = restr_obj.get_unsatisfied(data, apply_to_all)

    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i, j]
            ax.grid(False)
            ax.set_xticks([])
            ax.set_yticks([])

            if i * ncols + j < nattrs:
                attr = attrs[i * ncols + j]
                alpha_value = alpha[i * ncols + j] if isinstance(alpha, list) else alpha

                if is_numeric_dtype(data.obs[attr]):
                    values = data.obs[attr].values
                    assert apply_to_all or (not restr_obj.contains(attr))
                    values[unsel] = 0.0

                    img = ax.scatter(
                        df.iloc[:, 0],
                        df.iloc[:, 1],
                        c=values,
                        s=marker_size,
                        marker=".",
                        alpha=alpha_value,
                        edgecolors="none",
                        cmap="viridis",
                        rasterized=True,
                    )
                    img.set_clim(vmin, vmax)
                    left, bottom, width, height = ax.get_position().bounds
                    rect = [left + width * (1.0 + 0.05), bottom, width * 0.1, height]
                    ax_colorbar = fig.add_axes(rect)
                    fig.colorbar(img, cax=ax_colorbar)

                else:
                    labels = data.obs[attr].astype(str)
                    if (not apply_to_all) and restr_obj.contains(attr):
                        idx = restr_obj.get_unsatisfied_per_attr(labels, attr)
                    else:
                        idx = unsel
                    labels[idx] = ""
                    labels = as_category(labels)
                    label_size = labels.categories.size

                    if others["palettes"] is not None:
                        palettes = np.array(others["palettes"].split(","))
                    else:
                        palettes = get_palettes(
                            label_size,
                            with_background=idx.sum() > 0,
                            show_background=show_background,
                        )

                    for k, cat in enumerate(labels.categories):
                        idx = np.isin(labels, cat)
                        ax.scatter(
                            df.iloc[idx, 0],
                            df.iloc[idx, 1],
                            c=palettes[k],
                            s=marker_size,
                            marker=".",
                            alpha=alpha_value,
                            edgecolors="none",
                            label=cat,
                            rasterized=True,
                        )

                    legend = ax.legend(
                        loc="center left",
                        bbox_to_anchor=(1, 0.5),
                        frameon=False,
                        fontsize=legend_fontsize,
                        ncol=get_legend_ncol(label_size),
                    )
                    for handle in legend.legendHandles:
                        handle.set_sizes([300.0])

                ax.set_title(attr)
            else:
                ax.set_frame_on(False)

            if i == nrows - 1:
                ax.set_xlabel(basis + "1")
            if j == 0:
                ax.set_ylabel(basis + "2")

    return fig


### Sample usage:
###    fig = plot_scatter_groups(data, 'tsne', 'louvain_labels', 'Individual', nrows = 2, ncols = 4, alpha = 0.5)
def plot_scatter_groups(
    data,
    basis,
    cluster,
    group,
    restrictions=[],
    nrows=None,
    ncols=None,
    subplot_size=(4, 4),
    left=None,
    bottom=None,
    wspace=None,
    hspace=None,
    alpha=None,
    legend_fontsize=None,
    showall=True,
    **others,
):
    df = pd.DataFrame(
        data.obsm["X_" + basis][:, 0:2], columns=[basis + c for c in ["1", "2"]]
    )
    basis = transform_basis(basis)

    marker_size = get_marker_size(df.shape[0])

    labels = as_category(data.obs[cluster])
    label_size = labels.categories.size
    palettes = get_palettes(label_size)
    legend_ncol = get_legend_ncol(label_size)

    assert group in data.obs
    groups = as_category(data.obs[group])
    df_g = pd.DataFrame()

    if showall:
        df_g["All"] = np.ones(data.shape[0], dtype=bool)
    if len(restrictions) == 0:
        for cat in groups.categories:
            df_g[cat] = np.isin(groups, cat)
    else:
        restr_obj = RestrictionParser(restrictions)
        for key in restr_obj.get_attrs():
            df_g[key] = restr_ojb.get_satisfied_per_attr(groups, key)

    nrows, ncols = get_nrows_and_ncols(df_g.shape[1], nrows, ncols)

    kwargs = set_up_kwargs(subplot_size, left, bottom, wspace, hspace)
    fig, axes = get_subplot_layouts(nrows=nrows, ncols=ncols, squeeze=False, **kwargs)

    if legend_fontsize is None:
        legend_fontsize = rcParams["legend.fontsize"]

    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i, j]
            ax.grid(False)
            ax.set_xticks([])
            ax.set_yticks([])

            gid = i * ncols + j
            if gid < df_g.shape[1]:
                for k, cat in enumerate(labels.categories):
                    idx = np.logical_and(df_g.iloc[:, gid].values, np.isin(labels, cat))
                    ax.scatter(
                        df.iloc[idx, 0],
                        df.iloc[idx, 1],
                        c=palettes[k],
                        s=marker_size,
                        marker=".",
                        alpha=alpha,
                        edgecolors="none",
                        label=cat,
                        rasterized=True,
                    )

                ax.set_title(df_g.columns[gid])
                legend = ax.legend(
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                    frameon=False,
                    fontsize=legend_fontsize,
                    ncol=legend_ncol,
                )

                for handle in legend.legendHandles:
                    handle.set_sizes([300.0])
            else:
                ax.set_frame_on(False)

            if i == nrows - 1:
                ax.set_xlabel(basis + "1")
            if j == 0:
                ax.set_ylabel(basis + "2")

    return fig


### Sample usage:
###    fig = plot_scatter_genes(data, 'tsne', ['CD8A', 'CD4', 'CD3G', 'MS4A1', 'NCAM1', 'CD14', 'ITGAX', 'IL3RA', 'CD38', 'CD34', 'PPBP'])
def plot_scatter_genes(
    data,
    basis,
    genes,
    nrows=None,
    ncols=None,
    subplot_size=(4, 4),
    left=None,
    bottom=None,
    wspace=0.3,
    hspace=None,
    alpha=None,
    use_raw=False,
    **others,
):
    df = pd.DataFrame(
        data.obsm["X_" + basis][:, 0:2], columns=[basis + c for c in ["1", "2"]]
    )
    basis = transform_basis(basis)

    ngenes = len(genes)
    nrows, ncols = get_nrows_and_ncols(ngenes, nrows, ncols)
    marker_size = 240000.0 / df.shape[0]

    kwargs = set_up_kwargs(subplot_size, left, bottom, wspace, hspace)
    fig, axes = get_subplot_layouts(nrows=nrows, ncols=ncols, squeeze=False, **kwargs)

    X = data.raw[:, genes].X.toarray() if use_raw else data[:, genes].X.toarray()

    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i, j]
            ax.grid(False)
            ax.set_xticks([])
            ax.set_yticks([])

            if i * ncols + j < ngenes:
                gene_id = i * ncols + j
                img = ax.scatter(
                    df.iloc[:, 0],
                    df.iloc[:, 1],
                    s=marker_size,
                    c=X[:, gene_id],
                    marker=".",
                    alpha=alpha,
                    edgecolors="none",
                    cmap="viridis",
                    rasterized=True,
                )

                left, bottom, width, height = ax.get_position().bounds
                rect = [left + width * (1.0 + 0.05), bottom, width * 0.1, height]
                ax_colorbar = fig.add_axes(rect)
                fig.colorbar(img, cax=ax_colorbar)

                ax.set_title(genes[gene_id])
            else:
                ax.set_frame_on(False)

            if i == nrows - 1:
                ax.set_xlabel(basis + "1")
            if j == 0:
                ax.set_ylabel(basis + "2")

    return fig


### Sample usage:
###    fig = plot_scatter_gene_groups(data, 'tsne', 'CD8A', 'Individual', nrows = 3, ncols = 3)
def plot_scatter_gene_groups(
    data,
    basis,
    gene,
    group,
    nrows=None,
    ncols=None,
    subplot_size=(4, 4),
    left=None,
    bottom=None,
    wspace=0.3,
    hspace=None,
    alpha=None,
    use_raw=False,
    **others,
):
    df = pd.DataFrame(
        data.obsm["X_" + basis][:, 0:2], columns=[basis + c for c in ["1", "2"]]
    )
    basis = transform_basis(basis)

    marker_size = 240000.0 / df.shape[0]
    groups = as_category(data.obs[group])
    ngroup = groups.categories.size
    nrows, ncols = get_nrows_and_ncols(ngroup + 1, nrows, ncols)

    kwargs = set_up_kwargs(subplot_size, left, bottom, wspace, hspace)
    fig, axes = get_subplot_layouts(nrows=nrows, ncols=ncols, squeeze=False, **kwargs)

    X = (data.raw[:, gene].X if use_raw else data[:, gene].X).toarray()[:, 0]
    vmax = X.max()

    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i, j]
            ax.grid(False)
            ax.set_xticks([])
            ax.set_yticks([])

            gid = i * ncols + j
            if gid <= ngroup:
                if gid == 0:
                    idx_g = np.ones(groups.shape[0], dtype=bool)
                else:
                    idx_g = np.isin(groups, groups.categories[gid - 1])

                img = ax.scatter(
                    df.iloc[idx_g, 0],
                    df.iloc[idx_g, 1],
                    s=marker_size,
                    c=X[idx_g],
                    marker=".",
                    alpha=alpha,
                    edgecolors="none",
                    cmap="viridis",
                    vmin=0.0,
                    vmax=vmax,
                    rasterized=True,
                )

                left, bottom, width, height = ax.get_position().bounds
                rect = [left + width * (1.0 + 0.05), bottom, width * 0.1, height]
                ax_colorbar = fig.add_axes(rect)
                fig.colorbar(img, cax=ax_colorbar)

                ax.set_title("All" if gid == 0 else str(groups.categories[gid - 1]))
            else:
                ax.set_frame_on(False)

            if i == nrows - 1:
                ax.set_xlabel(basis + "1")
            if j == 0:
                ax.set_ylabel(basis + "2")

    return fig


### Sample usage:
###     cg = plot_heatmap(data, 'louvain_labels', ['CD8A', 'CD4', 'CD3G', 'MS4A1', 'NCAM1', 'CD14', 'ITGAX', 'IL3RA', 'CD38', 'CD34', 'PPBP'], use_raw = True, title="markers")
###     cg.savefig("heatmap.png", bbox_inches='tight', dpi=600)
def plot_heatmap(
    data, cluster, genes, use_raw=False, showzscore=False, title="", **kwargs
):
    sns.set(font_scale=0.35)

    adata = data.raw if use_raw else data
    df = pd.DataFrame(adata[:, genes].X.toarray(), index=data.obs.index, columns=genes)
    if showzscore:
        df = df.apply(zscore, axis=0)

    cluster_ids = as_category(data.obs[cluster])
    idx = cluster_ids.argsort()
    df = df.iloc[idx, :]  # organize df by category order
    row_colors = np.zeros(df.shape[0], dtype=object)
    palettes = get_palettes(cluster_ids.categories.size)

    cluster_ids = cluster_ids[idx]
    for k, cat in enumerate(cluster_ids.categories):
        row_colors[np.isin(cluster_ids, cat)] = palettes[k]

    cg = sns.clustermap(
        data=df,
        cmap="RdBu_r",
        row_colors=row_colors,
        row_cluster=False,
        col_cluster=True,
        linewidths=0,
        yticklabels=[],
        xticklabels=genes,
        **kwargs
    )
    cg.ax_heatmap.set_ylabel("")
    # move the colorbar
    cg.ax_row_dendrogram.set_visible(False)
    dendro_box = cg.ax_row_dendrogram.get_position()
    dendro_box.x0 = (dendro_box.x0 + 2 * dendro_box.x1) / 3
    dendro_box.x1 = dendro_box.x0 + 0.02
    cg.cax.set_position(dendro_box)
    cg.cax.yaxis.set_ticks_position("left")
    cg.cax.tick_params(labelsize=10)
    # draw a legend for the cluster groups
    cg.ax_col_dendrogram.clear()
    for k, cat in enumerate(cluster_ids.categories):
        cg.ax_col_dendrogram.bar(0, 0, color=palettes[k], label=cat, linewidth=0)
    cg.ax_col_dendrogram.legend(loc="center", ncol=15, fontsize=10)
    cg.ax_col_dendrogram.grid(False)
    cg.ax_col_dendrogram.set_xticks([])
    cg.ax_col_dendrogram.set_yticks([])

    return cg


### Sample usage:
###     cg = plot_violin_genes(data, 'louvain_labels', ['CD8A', 'CD4', 'CD3G', 'MS4A1', 'NCAM1', 'CD14', 'ITGAX', 'IL3RA', 'CD38', 'CD34', 'PPBP'], use_raw = True, title="markers")
###     cg.savefig("heatmap.png", bbox_inches='tight', dpi=600)
def plot_violin_genes(data, cluster, genes, subplot_size, ylab):
    nrows, ncols = get_nrows_and_ncols(len(genes), None, None)
    fig, axes = get_subplot_layouts(
        nrows=nrows, ncols=ncols, subplot_size=subplot_size, hspace=0.3, wspace=0.1, squeeze=False
    )
    expr_mat = data[:, genes].X.toarray()
    df = pd.DataFrame(data=expr_mat, columns=genes)
    df.insert(0, "label", data.obs[cluster].values)
    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i, j]
            idx = i * ncols + j
            if idx < len(genes):
                sns.violinplot(
                    x="label", y=genes[idx], data=df, inner=None, linewidth=0, ax=ax, cut=0
                )
                sns.stripplot(x="label", y=genes[idx], data=df, ax=ax, size=2, color="k")
                ax.set_xlabel("")
                ax.set_ylabel("")
                ax.set_title(genes[idx])
            else:
                ax.set_frame_on(False)
    plt.figtext(0.02, 0.5, ylab, rotation="vertical", fontsize="xx-large")
    return fig



import time
import numpy as np
import pandas as pd

from scipy.sparse import issparse
from collections import defaultdict
from joblib import Parallel, delayed
import skmisc.loess as sl
from typing import List
from anndata import AnnData

import logging
logger = logging.getLogger("pegasus")
from pegasus.utils import decorators as pg_deco


def estimate_feature_statistics(data: AnnData, consider_batch: bool) -> None:
    """ Estimate feature (gene) statistics per channel, such as mean, var etc.
    """
    assert issparse(data.X)

    if consider_batch:
        start = time.perf_counter()
        if "Channels" not in data.uns:
            data.uns["Channels"] = data.obs["Channel"].unique()

        if "Group" not in data.obs:
            data.obs["Group"] = "one_group"

        if "Groups" not in data.uns:
            data.uns["Groups"] = data.obs["Group"].unique()

        channels = data.uns["Channels"]
        groups = data.uns["Groups"]

        ncells = np.zeros(channels.size)
        means = np.zeros((data.shape[1], channels.size))
        partial_sum = np.zeros((data.shape[1], channels.size))

        group_dict = defaultdict(list)
        for i, channel in enumerate(channels):
            idx = np.isin(data.obs["Channel"], channel)
            mat = data.X[idx].astype(np.float64)
            ncells[i] = mat.shape[0]

            if ncells[i] == 0:
                continue

            if ncells[i] == 1:
                means[:, i] = mat.toarray()[0]
            else:
                means[:, i] = mat.mean(axis=0).A1
                m2 = mat.power(2).sum(axis=0).A1
                partial_sum[:, i] = m2 - ncells[i] * (means[:, i] ** 2)

            group = data.obs["Group"][idx.nonzero()[0][0]]
            group_dict[group].append(i)

        partial_sum[partial_sum < 1e-6] = 0.0

        overall_means = np.dot(means, ncells) / data.shape[0]
        batch_adjusted_vars = np.zeros(data.shape[1])

        c2gid = np.zeros(channels.size, dtype=int)
        gncells = np.zeros(groups.size)
        gmeans = np.zeros((data.shape[1], groups.size))
        gstds = np.zeros((data.shape[1], groups.size))

        for i, group in enumerate(groups):
            gchannels = group_dict[group]
            c2gid[gchannels] = i
            gncells[i] = ncells[gchannels].sum()
            gmeans[:, i] = np.dot(means[:, gchannels], ncells[gchannels]) / gncells[i]
            gstds[:, i] = (
                partial_sum[:, gchannels].sum(axis=1) / gncells[i]
            ) ** 0.5  # calculate std
            if groups.size > 1:
                batch_adjusted_vars += gncells[i] * (
                    (gmeans[:, i] - overall_means) ** 2
                )

        data.varm["means"] = means
        data.varm["partial_sum"] = partial_sum
        data.uns["ncells"] = ncells

        data.varm["gmeans"] = gmeans
        data.varm["gstds"] = gstds
        data.uns["gncells"] = gncells
        data.uns["c2gid"] = c2gid

        data.var["mean"] = overall_means
        data.var["var"] = (batch_adjusted_vars + partial_sum.sum(axis=1)) / (
            data.shape[0] - 1.0
        )
        end = time.perf_counter()
        logger.info(
            "Estimation on feature statistics per channel is finished. Time spent = {:.2f}s.".format(
                end - start
            )
        )
    else:
        mean = data.X.mean(axis=0).A1
        m2 = data.X.power(2).sum(axis=0).A1
        var = (m2 - data.X.shape[0] * (mean ** 2)) / (data.X.shape[0] - 1)

        data.var["mean"] = mean
        data.var["var"] = var


def fit_loess(x: List[float], y: List[float], span: float, degree: int) -> object:
    try:
        lobj = sl.loess(x, y, span=span, degree=2)
        lobj.fit()
        return lobj
    except ValueError:
        return None


def select_hvf_pegasus(
    data: AnnData, consider_batch: bool, n_top: int = 2000, span: float = 0.02
) -> None:
    """ Select highly variable features using the pegasus method
    """
    if "robust" not in data.var:
        raise ValueError("Please run `qc_metrics` to identify robust genes")

    estimate_feature_statistics(data, consider_batch)

    robust_idx = data.var["robust"].values
    hvf_index = np.zeros(robust_idx.sum(), dtype=bool)

    mean = data.var.loc[robust_idx, "mean"]
    var = data.var.loc[robust_idx, "var"]

    span_value = span
    while True:
        lobj = fit_loess(mean, var, span = span_value, degree = 2)
        if lobj is not None:
            break
        span_value += 0.01
    if span_value > span:
        logger.warning("Leoss span is adjusted from {:.2f} to {:.2f} to avoid fitting errors.".format(span, span_value))

    rank1 = np.zeros(hvf_index.size, dtype=int)
    rank2 = np.zeros(hvf_index.size, dtype=int)

    delta = var - lobj.outputs.fitted_values
    fc = var / lobj.outputs.fitted_values

    rank1[np.argsort(delta)[::-1]] = range(hvf_index.size)
    rank2[np.argsort(fc)[::-1]] = range(hvf_index.size)
    hvf_rank = rank1 + rank2

    hvf_index[np.argsort(hvf_rank)[:n_top]] = True

    data.var["hvf_loess"] = 0.0
    data.var.loc[robust_idx, "hvf_loess"] = lobj.outputs.fitted_values

    data.var["hvf_rank"] = -1
    data.var.loc[robust_idx, "hvf_rank"] = hvf_rank
    data.var["highly_variable_features"] = False
    data.var.loc[robust_idx, "highly_variable_features"] = hvf_index


def select_hvf_seurat_single(
    X: "csr_matrix",
    n_top: int,
    min_disp: float,
    max_disp: float,
    min_mean: float,
    max_mean: float,
) -> List[int]:
    """ HVF selection for one channel using Seurat method
    """
    X = X.copy().expm1()
    mean = X.mean(axis=0).A1
    m2 = X.power(2).sum(axis=0).A1
    var = (m2 - X.shape[0] * (mean ** 2)) / (X.shape[0] - 1)

    dispersion = np.full(X.shape[1], np.nan)
    idx_valid = (mean > 0.0) & (var > 0.0)
    dispersion[idx_valid] = var[idx_valid] / mean[idx_valid]

    mean = np.log1p(mean)
    dispersion = np.log(dispersion)

    df = pd.DataFrame({"log_dispersion": dispersion, "bin": pd.cut(mean, bins=20)})
    log_disp_groups = df.groupby("bin")["log_dispersion"]
    log_disp_mean = log_disp_groups.mean()
    log_disp_std = log_disp_groups.std(ddof=1)
    log_disp_zscore = (
        df["log_dispersion"].values - log_disp_mean.loc[df["bin"]].values
    ) / log_disp_std.loc[df["bin"]].values
    log_disp_zscore[np.isnan(log_disp_zscore)] = 0.0

    hvf_rank = np.full(X.shape[1], -1, dtype=int)
    ords = np.argsort(log_disp_zscore)[::-1]

    if n_top is None:
        hvf_rank[ords] = range(X.shape[1])
        idx = np.logical_and.reduce(
            (
                mean > min_mean,
                mean < max_mean,
                log_disp_zscore > min_disp,
                log_disp_zscore < max_disp,
            )
        )
        hvf_rank[~idx] = -1
    else:
        hvf_rank[ords[:n_top]] = range(n_top)

    return hvf_rank


def select_hvf_seurat_multi(
    X: "csr_matrix",
    channels: List[str],
    cell2channel: List[str],
    n_top: int,
    n_jobs: int,
    min_disp: float,
    max_disp: float,
    min_mean: float,
    max_mean: float,
) -> List[int]:
    Xs = []
    for channel in channels:
        Xs.append(X[np.isin(cell2channel, channel)])

    from joblib import effective_n_jobs

    n_jobs = effective_n_jobs(n_jobs)

    res_arr = np.array(
        Parallel(n_jobs=n_jobs)(
            delayed(select_hvf_seurat_single)(
                Xs[i], n_top, min_disp, max_disp, min_mean, max_mean
            )
            for i in range(channels.size)
        )
    )
    selected = res_arr >= 0
    shared = selected.sum(axis=0)
    cands = (shared > 0).nonzero()[0]
    import numpy.ma as ma

    median_rank = ma.median(ma.masked_array(res_arr, mask=~selected), axis=0).data
    cands = sorted(cands, key=lambda x: median_rank[x])
    cands = sorted(cands, key=lambda x: shared[x], reverse=True)

    hvf_rank = np.full(X.shape[1], -1, dtype=int)
    hvf_rank[cands[:n_top]] = range(n_top)

    return hvf_rank


def select_hvf_seurat(
    data: AnnData,
    consider_batch: bool,
    n_top: int,
    min_disp: float,
    max_disp: float,
    min_mean: float,
    max_mean: float,
    n_jobs: int,
) -> None:
    """ Select highly variable features using Seurat method.
    """

    robust_idx = data.var["robust"].values
    X = data.X[:, robust_idx]

    hvf_rank = (
        select_hvf_seurat_multi(
            X,
            data.uns["Channels"],
            data.obs["Channel"],
            n_top,
            n_jobs=n_jobs,
            min_disp=min_disp,
            max_disp=max_disp,
            min_mean=min_mean,
            max_mean=max_mean,
        )
        if consider_batch
        else select_hvf_seurat_single(
            X,
            n_top=n_top,
            min_disp=min_disp,
            max_disp=max_disp,
            min_mean=min_mean,
            max_mean=max_mean,
        )
    )

    hvf_index = hvf_rank >= 0

    data.var["hvf_rank"] = -1
    data.var.loc[robust_idx, "hvf_rank"] = hvf_rank
    data.var["highly_variable_features"] = False
    data.var.loc[robust_idx, "highly_variable_features"] = hvf_index

@pg_deco.TimeLogger()
def highly_variable_features(
    data: AnnData,
    consider_batch: bool,
    flavor: str = "pegasus",
    n_top: int = 2000,
    span: float = 0.02,
    min_disp: float = 0.5,
    max_disp: float = np.inf,
    min_mean: float = 0.0125,
    max_mean: float = 7,
    n_jobs: int = -1,
) -> None:
    """ Highly variable features (HVF) selection. The input data should be logarithmized.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    consider_batch: ``bool``.
        Whether consider batch effects or not.

    flavor: ``str``, optional, default: ``"pegasus"``
        The HVF selection method to use. Available choices are ``"pegasus"`` or ``"Seurat"``.

    n_top: ``int``, optional, default: ``2000``
        Number of genes to be selected as HVF. if ``None``, no gene will be selected.

    span: ``float``, optional, default: ``0.02``
        Only applicable when ``flavor`` is ``"pegasus"``. The smoothing factor used by *scikit-learn loess* model in pegasus HVF selection method.

    min_disp: ``float``, optional, default: ``0.5``
        Minimum normalized dispersion.

    max_disp: ``float``, optional, default: ``np.inf``
        Maximum normalized dispersion. Set it to ``np.inf`` for infinity bound.

    min_mean: ``float``, optional, default: ``0.0125``
        Minimum mean.

    max_mean: ``float``, optional, default: ``7``
        Maximum mean.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to be used during calculation. If ``-1``, all available threads will be used.

    Returns
    -------

    Examples
    --------
    >>> pg.highly_variable_features(adata, consider_batch = False)
    """

    if "Channels" not in data.uns:
        if "Channel" not in data.obs:
            data.obs["Channel"] = ""
        data.uns["Channels"] = data.obs["Channel"].unique()

    if data.uns["Channels"].size == 1 and consider_batch:
        consider_batch = False
        logger.warning(
            "Warning: only contains one channel, no need to consider batch for selecting highly variable features."
        )

    if flavor == "pegasus":
        select_hvf_pegasus(data, consider_batch, n_top=n_top, span=span)
    else:
        assert flavor == "Seurat"
        select_hvf_seurat(
            data,
            consider_batch,
            n_top=n_top,
            min_disp=min_disp,
            max_disp=max_disp,
            min_mean=min_mean,
            max_mean=max_mean,
            n_jobs=n_jobs,
        )

    logger.info(
        "{} highly variable features have been selected.".format(
            data.var["highly_variable_features"].sum()
        )
    )
